"""Pipeline sections that filter the incoming items."""
from async_generator import aclosing

from .abc import Section

class Skip(Section):
    """Skips the first count items in an asynchronous sequence.

    Skip can be used as a starting section if a source is given.

    Args:
        count (int): Number of items to skip
        source (AsyncIterable[Any]): Input source if starting section.
    """
    def __init__(self, count: int, source: None):
        super().__init__()
        self.count = count
        self.source = source

    async def pump(self, input, output):
        if input is None:
            if self.source is not None:
                input = self.source
            else:
                raise RuntimeError('No input provided.')
        async with aclosing(input) as aiter, output:
            for _ in range(self.count):
                await aiter.__anext__()
            async for item in aiter:
                await output.send(item)

class Filter(Section):
    """Outputs items that match a filter.

    The filter function must take an item. If the return value evaluates as true, the item is sent,
    otherwise the item is discarded.

    Filter can be used as a starting section, if a source is provided.

    Args:
        func (Callable[[Any], bool]): Matching function.
        source (AsyncIterable[Any]): Source if used as a starting section.
    """
    def __init__(self, func, source=None):
        super().__init__()
        self.func = func
        self.source = source

    async def pump(self, input, output):
        if input is None:
            if self.source is not None:
                input = self.source
            else:
                raise RuntimeError('No input provided.')

        async with aclosing(input) as aiter, output:
            async for item in aiter:
                if self.func(item):
                    await output.send(item)

class Changes(Section):
    """Outputs items that are different from the last item output.

    The generator stores a reference to the last output item. Whenever a new item arrives, it is
    compared to the previously output item. If they are equal, the new item is discarded. If not,
    the new item is output and becomes the new reference. The first item received is always
    outputted.

    Changes can be used as a starting section, if a source is provided.

    Note: Items are compared using the != operator.

    Args:
        source (AsyncIterable[Any]): Source if used as a starting section.
    """
    def __init__(self, source=None):
        super().__init__()
        self.source = source

    async def pump(self, input, output):
        if input is None:
            if self.source is not None:
                input = self.source
            else:
                raise RuntimeError('No input provided.')
        token = object()
        last = token
        async with aclosing(input) as aiter, output:
            async for item in aiter:
                if last is token or item != last:
                    last = item
                    await output.send(item)



"""Sections for transforming an input into a different output."""
from async_generator import aclosing

from .abc import Section

class Map(Section):
    """Maps over an asynchronous sequence.

    Map can be used as a starting section, if a source is provided.

    Args:
        func (Callable[[Any], Any]): Mapping function.
        source (AsyncIterable[Any]): Source if used as a starting section.
    """
    def __init__(self, func, source=None):
        super().__init__()
        self.func = func
        self.source = source

    async def pump(self, input, output):
        if input is None:
            if self.source is not None:
                input = self.source
            else:
                raise RuntimeError('No input provided.')
        async with aclosing(input) as aiter, output:
            async for item in aiter:
                await output.send(self.func(item))



"""Main Pipeline class"""
__version__ = '0.1.0'

from contextlib import asynccontextmanager
from itertools import chain
from typing import Sequence

import trio
from async_generator import aclosing

from .abc import Section
from .tap import Tap

class Pipeline:
    def __init__(self, *sections: Sequence[Section],
                 main_nursery: trio.Nursery,
                 main_switch: trio.Event):
        self.sections = sections
        self._taps = set()
        self._main_nursery = main_nursery
        self._main_switch = main_switch

    @classmethod
    @asynccontextmanager
    async def create(cls, *sections: Sequence[Section]):
        async with trio.open_nursery() as nursery:
            pipeline = cls(*sections, main_nursery=nursery, main_switch=trio.Event())
            nursery.start_soon(pipeline._pump) # pylint: disable=protected-access
            yield pipeline
            nursery.cancel_scope.cancel()

    async def _pump(self):
        await self._main_switch.wait()

        if isinstance(self.sections[0], Section):
            first_input = None
            sections = self.sections
        else:
            first_input = self.sections[0]
            sections = self.sections[1:]

        channels = chain((first_input,), *(trio.open_memory_channel(0) for _ in sections))

        async with trio.open_nursery() as nursery:

            # Start pumps
            for section in sections:
                nursery.start_soon(section.pump, next(channels), next(channels))

            # Output to taps
            async with aclosing(next(channels)) as aiter:
                async for item in aiter:
                    for tap in self._taps:
                        nursery.start_soon(tap.send, item)

        # There is no more output to send. Close the taps.
        for tap in self._taps:
            await tap.send_channel.aclose()

    def tap(self, *,
            max_buffer_size: int = 0,
            timeout: float = 1,
            retrys: int = 3,
            start: bool = True) -> trio.MemoryReceiveChannel:
        """Create a new output channel for this pipeline."""
        send_channel, receive_channel = trio.open_memory_channel(max_buffer_size)
        self._taps.add(Tap(send_channel, timeout, retrys))
        if start:
            self._main_switch.set()
        return receive_channel

    def extend(self, *sections: Sequence[Section]) -> "Pipeline":
        """Extend this pipeline into a new pipeline."""
        pipeline = Pipeline(
            self.tap(),
            sections,
            main_nursery=self._main_nursery,
            main_switch=self._main_switch
        )
        self._main_nursery.start_soon(pipeline._pump) # pylint: disable=protected-access
        return pipeline



""" Abstract Base Classes for building pipeline sections. """
from abc import ABC, abstractmethod
from typing import Any, AsyncIterable, Union

import trio

class Section(ABC):
    """Each pipeline section takes inputs from an async iterable, processes it and sends it to an
    output.
    """

    @abstractmethod
    async def pump(self, input: Union[AsyncIterable[Any], None], output: trio.MemorySendChannel):
        """The run method must contain the logic that iterates the input, processes the indidual
        items, and feeds results to the output.

        The first section of the pipeline does not receive an input. It is expected to supply
        itself.
        """



__version__ = '0.2.0'

from .pipeline import Pipeline

from .abc import Section
from .buffers import FiloBuffer, Group, Delay, RateLimit
from .combiners import Chain, Merge, Zip, ZipLatest
from .filters import Skip, Filter, Changes
from .refiners import Map



"""Miscelaneous generators that doesn't fit the Mixer-Refiner pattern.

Can be used as pipeline input."""
import trio

async def repeat(interval, item):
    """Yields a single item repeatedly at regular intervals."""
    while True:
        yield item
        await trio.sleep(interval)



import trio

class Tap:
    def __init__(self, send_channel, timeout, retrys):
        self.send_channel = send_channel
        self.timeout = timeout
        self.retrys = retrys

    async def send(self, item):
        for _ in range(self.retrys + 1):
            with trio.move_on_after(self.timeout):
                await self.send_channel.send(item)
                return
            await trio.sleep(0)
        raise trio.BusyResourceError('Unable to send item.')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slurry']

package_data = \
{'': ['*']}

install_requires = \
['trio>=0.16.0,<0.17.0']

setup_kwargs = {
    'name': 'slurry',
    'version': '0.2.0',
    'description': 'An async stream processing microframework',
    'long_description': '.. image:: https://img.shields.io/pypi/v/slurry.svg\n   :target: https://pypi.org/project/slurry\n   :alt: Latest PyPi version\n\n.. image:: https://travis-ci.com/andersea/slurry.svg?branch=master\n   :target: https://travis-ci.com/andersea/slurry\n   :alt: Build Status\n\nSlurry - An async stream processing microframework\n==================================================',
    'author': 'Anders Ellenshøj Andersen',
    'author_email': 'andersa@ellenshoej.dk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andersea/slurry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Pipeline sections with age- and volume-based buffers."""
from collections import deque
import math
from typing import Any, Callable, Sequence

import trio
from async_generator import aclosing

from .abc import Section

class FiloBuffer(Section):
    """First in-last out sequence buffer with optional.

    Iterates an asynchronous sequence and puts each received item in a buffer..
    When an item is received, the buffer is filtered by dumping the oldest items first, until the
    size of the buffer is equal to max_size or less. The buffer is then output as a tuple.

    If an interval is given, items have a time limit, after which the item is discarded, when
    filtering occurs.

    It is possible to set a minimum buffer size. If the buffer is shorter than min_size, items will
    be added, but the buffer will not be yielded.

    All items remain in the buffer after they are yielded, and can be yielded again, unless they
    are removed by one of the filter conditions."""
    def __init__(self, max_size, source=None, *, max_age=math.inf, min_size=1):
        super().__init__()
        self.source = source
        self.max_size = max_size
        self.max_age = max_age
        self.min_size = min_size

    async def pump(self, input, output):
        if self.source is not None:
            input = self.source
        buf = deque()
        async with aclosing(input) as aiter, output:
            async for item in aiter:
                now = trio.current_time()
                buf.append((item, now))
                while len(buf) > self.max_size or now - buf[0][1] > self.max_age:
                    buf.popleft()
                if len(buf) >= self.min_size:
                    await output.send(tuple(i[0] for i in buf))

class Group(Section):
    """Groups received items by time based interval.

    Group awaits an item to arrive from source, adds it to a buffer and sets a timeout using
    interval. While the timeout is active, additional items received are added to the buffer.
    When the timeout triggers, or if the buffer size equals max_size, the buffer is yielded and
    a new empty buffer is created.

    The buffer is not yielded at regular intervals. The timeout only starts if there are items in
    the buffer, thus the minimum length of the buffer, when it is yielded, is always 1.

    The items in the buffer can optionally be mapped over, by supplying a mapper function.

    The items can be reduced to a single value, by supplying a reducer function. Note, that the
    reducer function must take the entire buffer as a single argument and return a single value."""

    def __init__(self, interval, source=None, *,
                 max_size=math.inf,
                 mapper: Callable[[Any], Any] = None,
                 reducer: Callable[[Sequence[Any]], Any] = None):
        super().__init__()
        self.source = source
        self.interval = interval
        self.max_size = max_size
        self.mapper = mapper
        self.reducer = reducer

    async def pump(self, input, output):
        if input is None:
            if self.source is not None:
                input = self.source
            else:
                raise RuntimeError('No input provided.')
        async with output, aclosing(input) as aiter:
            while True:
                buffer = []
                try:
                    buffer.append(await aiter.__anext__())
                    with trio.move_on_after(self.interval):
                        while True:
                            if len(buffer) == self.max_size:
                                break
                            buffer.append(await aiter.__anext__())
                except StopAsyncIteration:
                    if buffer:
                        await output.send(self._process_result(buffer))
                    break
                else:
                    await output.send(self._process_result(buffer))

    def _process_result(self, buffer):
        if self.mapper is not None:
            buffer = (self.mapper(x) for x in buffer)
        if self.reducer is not None:
            return self.reducer(buffer)
        else:
            return tuple(buffer)

class Delay(Section):
    """Delays transmission of each item received by an interval.

    Received items are temporarily stored in an unbounded queue, along with a timestamp, using
    a background task. The foreground task takes items from the queue, and waits until the
    item is older than the given interval and then transmits it."""
    def __init__(self, interval: float, source=None):
        super().__init__()
        self.source = source
        self.interval = interval

    async def pump(self, input, output):
        if self.source is not None:
            input = self.source
        buffer_input_channel, buffer_output_channel = trio.open_memory_channel(math.inf)

        async def pull_task():
            async with buffer_input_channel, aclosing(input) as aiter:
                async for item in aiter:
                    await buffer_input_channel.send((item, trio.current_time() + self.interval))

        async with trio.open_nursery() as nursery:
            nursery.start_soon(pull_task)
            async with buffer_output_channel, output:
                async for item, timestamp in buffer_output_channel:
                    now = trio.current_time()
                    if timestamp > now:
                        await trio.sleep(timestamp - now)
                    await output.send(item)

class RateLimit(Section):
    """Limits data rate of an input to a certain interval.

    The first item received is transmitted and a timeout starts. Any other items received within
    interval time are discarded. When the interval is exceeded, the timeout resets and a new
    item can be transmitted."""
    def __init__(self, interval, source=None):
        super().__init__()
        self.source = source
        self.interval = interval

    async def pump(self, input, output):
        if self.source is not None:
            input = self.source
        then = 0
        async with aclosing(input) as aiter, output:
            async for item in aiter:
                now = trio.current_time()
                if now - then > self.interval:
                    then = now
                    await output.send(item)



"""Pipeline sections for combining multiple inputs into a single output."""
import builtins
import itertools
from typing import Sequence, AsyncIterable

import trio
from async_generator import aclosing

from .abc import Section

class Chain(Section):
    """Chains asynchronous sequences.

    Outputs items from each source in turn, until it is exhausted. If a source never reaches the
    end, remaining sources will not be iterated.

    Chain can be placed as a middle section and will chain the input of the previous section.

    Note:
        By default, the input is added as the first sources. If the input is added last instead
        of first, it will cause backpressure to be applied upstream.

    Args:
        *sources (AsyncIterable[Any]): One or more async iterables that will be chained together.
        place_input (string): Options: 'first' (default)|'last'. """
    def __init__(self, *sources, place_input='first'):
        super().__init__()
        self.sources = sources
        self.place_input = _validate_place_input(place_input)

    async def pump(self, input, output):
        if input is not None:
            if self.place_input == 'first':
                sources = (input, *self.sources)
            elif self.place_input == 'last':
                sources = (*self.sources, input)
        else:
            sources = self.sources
        async with output:
            for source in sources:
                async with aclosing(source) as agen:
                    async for item in agen:
                        await output.send(item)

class Merge(Section):
    """Merges asynchronous sequences.

    Sources are iterated in parallel and items are send from each source, as soon as they become
    available.

    If Merge is used as a middle section, the input will be added to the sources.

    Args:
        *sources (AsyncIterable[Any]): One or more async iterables that will be
            merged together.
    """
    def __init__(self, *sources):
        super().__init__()
        self.sources = sources

    async def pump(self, input, output):
        if input is not None:
            sources = (input, *self.sources)
        else:
            sources = self.sources
        async with output, trio.open_nursery() as nursery:

            async def pull_task(source, output):
                async with output, aclosing(source) as aiter:
                    async for item in aiter:
                        await output.send(item)

            for source in sources:
                nursery.start_soon(pull_task, source, output.clone())

class Zip(Section):
    """Zip asynchronous sequences.

    Sources are iterated in parallel and as soon as all sources have an item available, those
    items are output as a tuple.

    Zip can be used as a middle section, and the pipeline input will be added to the sources.

    Args:
        *sources (AsyncIterable[Any]): One or more async iterables that will be
            merged together.
        place_input (string):  Options: 'first' (default)|'last'.
    """
    def __init__(self, *sources: Sequence[AsyncIterable], place_input='first'):
        super().__init__()
        self.sources = sources
        self.place_input = _validate_place_input(place_input)

    async def pump(self, input, output):
        if input is not None:
            if self.place_input == 'first':
                sources = (input, *self.sources)
            elif self.place_input == 'last':
                sources = (*self.sources, input)
        else:
            sources = self.sources

        async with output:
            with trio.CancelScope() as cancel_scope:
                async def pull_task(source, results, index):
                    try:
                        results[index] = await source.__anext__()
                    except StopAsyncIteration:
                        cancel_scope.cancel()

                while True:
                    results = [None for _ in sources]
                    async with trio.open_nursery() as nursery:
                        for i, source in builtins.enumerate(sources):
                            nursery.start_soon(pull_task, source, results, i)
                    await output.send(tuple(results))

class ZipLatest(Section):
    """Zips asynchronous sequences and yields a result for on every received item.

    Sources are iterated in parallel and a tuple is yielded each time a result is ready
    on any source. The tuple values will be the last received value from each source.

    If any single source is exchausted, all remaining sources will be forcibly closed, and
    the generator will exit.

    Using monitor argument, one or more asynchronous sequences can be added, that will not trigger
    an output by themselves. Their latest value will be stored and added to the output value
    but will only be output if a new item arrives at one of the main sources.

    ZipLatest can be used as a middle section, in which case the upstream pipeline is
    added as an input.

    Args:
        *sources (AsyncIterable[Any]): One or more async iterables that will be
        zipped together.
        partial (bool): If True (default) output will be sent as soon as the first input arrives.
        default (Any): If partial is True, this is used as the default value when no input has
            arrived.
        monitor (Union[AsyncIterable[Any], Sequence[AsyncIterable[Any]]]): Asynchronous sequences to
            monitor.
        place_input (string): Options: 'first' (default)|'last'
        monitor_input (bool): Input is used as a monitored stream instead of a main source.
    """
    def __init__(self, *sources,
                 partial=True,
                 default=None,
                 monitor=(),
                 place_input='first',
                 monitor_input=False):
        super().__init__()
        self.sources = sources
        self.partial = partial
        self.default = default
        self.monitor = monitor
        self.place_input = _validate_place_input(place_input)
        self.monitor_input = monitor_input

    async def pump(self, input, output):
        sources = self.sources
        try:
            iter(self.monitor)
            monitor = self.monitor
        except TypeError:
            monitor = (self.monitor,)

        if input is not None:
            if self.monitor_input:
                if self.place_input == 'first':
                    monitor = (input, *monitor)
                elif self.place_input == 'last':
                    monitor = (*monitor, input)
            else:
                if self.place_input == 'first':
                    sources = (input, *sources)
                elif self.place_input == 'last':
                    sources = (*sources, input)

        results = [self.default for _ in itertools.chain(sources, monitor)]
        ready = [False for _ in results]

        async with output, trio.open_nursery() as nursery:

            async def pull_task(index, source, monitor=False):
                async with aclosing(source) as aiter:
                    async for item in aiter:
                        results[index] = item
                        ready[index] = True
                        if not monitor and (self.partial or False not in ready):
                            await output.send(tuple(results))
                nursery.cancel_scope.cancel()

            for i, source in builtins.enumerate(sources):
                nursery.start_soon(pull_task, i, source)
            for i, source in builtins.enumerate(monitor):
                nursery.start_soon(pull_task, i + len(sources), source, True)

def _validate_place_input(place_input):
    if isinstance(place_input, str):
        if place_input not in ['first', 'last']:
            raise ValueError(f'Invalid place_input argument: {place_input}')
    else:
        raise TypeError('place_input argument has invalid type.')
    return place_input
